/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.fa;

import java.util.Collections;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.CandidatePathList;
import com.net2plan.utils.Constants.OrderingType;
import com.net2plan.utils.Constants.SearchType;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.LongUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Set;

/**
 * Given a set of nodes and links, with their capacities, an offered traffic, and a set of admissible paths for each demand (given by the k-shortest paths
 * in km for each demand, being k a user-defined parameter), this algorithm uses an evolutionary algorithm metaheuristic to find for each demand two
 * link-disjoint routes: a primary route, and a backup route used for restoration. That is, the backup route is pre-computed, but it carries not traffic
 * unless the primary route fails. The optimization target is minimizing the average between the network congestion in the non-failure state, and the
 * worse network congestion among the single link failures. For each demand, the candidate link-disjoint candidate pairs are computed as all the
 * link disjoint pairs p1,p2, such that p1 and p2 are admissible paths. Two solutions are considered neighbors if they differ in the routing of one demand
 * (either primary path, backup path or both). A solution is codified as in int vector, with one coordinate per demand, containing the identifier of the
 * link-disjoint route pair for that demand. The population size, the number of generations and the offspring size are used-defined parameters.
 * In each generation, a user-defined fraction of the parents are selected randomly, while the rest are the top ones. Cross-over is uniform: for each demand
 * the routing of a randomly chosen parent passes to the child. Children are mutated in one demand randomly chosen. The top solutions among parents and
 * offspring make it to the next generation. The evolutionary core is implemented in a separated internal class, for didactic purposes.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class FA_EA_minCongestionPrecomputedRestoration implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		final int N = netPlan.getNumberOfNodes();
		final int D = netPlan.getNumberOfDemands();
		final int E = netPlan.getNumberOfLinks();
		if (N == 0 || E == 0 || D == 0) throw new Net2PlanException("This algorithm requires a topology with links and a demand set");

		/* Initialize some variables */
		final Map<Long, Double> u_e = netPlan.getLinkCapacityMap();
		final Map<Long, Double> h_d = netPlan.getDemandOfferedTrafficMap();
		final int k = Integer.parseInt(algorithmParameters.get("k")); /* number of loopless candidate paths per demand */
		final int ea_numberIterations = Integer.parseInt(algorithmParameters.get("ea_numberIterations"));
		final int ea_populationSize = Integer.parseInt(algorithmParameters.get("ea_populationSize"));
		final int ea_offSpringSize = Integer.parseInt(algorithmParameters.get("ea_offSpringSize"));
		final double ea_fractionParentsChosenRandomly = Double.parseDouble(algorithmParameters.get("ea_fractionParentsChosenRandomly"));
		if (ea_offSpringSize > ea_populationSize / 2.0) throw new Net2PlanException("The offspring size cannot exceed the population size divided by two");

		/* Construct the candidate path list */
		CandidatePathList cpl = new CandidatePathList(netPlan, netPlan.getLinkLengthInKmMap(), "K", Integer.toString(k));
		
		/* For each demand, compute those path pairs that are link disjoint */
		Map<Long, List<Pair<Long, Long>>> candidate11PairsList_d = computeCandidate11PairList(netPlan, cpl);

		EvolutionaryAlgorithmCore gac = new EvolutionaryAlgorithmCore(u_e, h_d, candidate11PairsList_d, cpl, ea_numberIterations, ea_populationSize, ea_fractionParentsChosenRandomly, ea_offSpringSize);
		gac.evolve();

		/* Save the best solution found into the netPlan object */
		netPlan.removeAllRoutes();
		netPlan.removeAllProtectionSegments();
		
		Set<Long> demandIds = netPlan.getDemandIds();
		for (long demandId : demandIds)
		{
			final int bestPairThisDemand = gac.best_solution.get(demandId);
			final Pair<Long, Long> primaryAndBackupPath = candidate11PairsList_d.get(demandId).get(bestPairThisDemand);
			final List<Long> primarySeqLinks = cpl.getPathSequenceOfLinks(primaryAndBackupPath.getFirst());
			final List<Long> backupSeqLinks = cpl.getPathSequenceOfLinks(primaryAndBackupPath.getSecond());

			final long segmentId = netPlan.addProtectionSegment(backupSeqLinks, 0, null);
			final long routeId = netPlan.addRoute(demandId, h_d.get(demandId), primarySeqLinks, null);
			netPlan.addProtectionSegmentToRouteBackupSegmentList(segmentId, routeId);
		}

		return "Ok! Cost : " + gac.best_cost;
	}

	@Override
	public String getDescription()
	{
		return "Given a set of nodes and links, with their capacities, an offered traffic, and a set of admissible paths for each demand (given by"
				+ " the k-shortest paths in km for each demand, being k a user-defined parameter), this algorithm uses an evolutionary algorithm "
				+ "metaheuristic to find for each demand two link-disjoint routes: a primary route, and a backup route used for restoration. That is, "
				+ "the backup route is pre-computed, but it carries not traffic unless the primary route fails. The optimization target is minimizing the average "
				+ "between the network congestion in the non-failure state, and the worse network congestion among the single link failures. For each demand, "
				+ "the candidate link-disjoint candidate pairs are computed as all the link disjoint pairs p1,p2, such that p1 and p2 are admissible paths. "
				+ "Two solutions are considered neighbors if they differ in the routing of one demand (either primary path, backup path or both). A solution is "
				+ "codified as in int vector, with one coordinate per demand, containing the identifier of the link-disjoint route pair for that demand. "
				+ "The population size, the number of generations and the offspring size are used-defined parameters. In each generation, a user-defined fraction"
				+ "of the parents are selected randomly, while the rest are the top ones. Cross-over is uniform: for each demand the routing of a randomly "
				+ "chosen parent passes to the child. Children are mutated in one demand randomly chosen. The top solutions among parents and offspring make it "
				+ "to the next generation. The evolutionary core is implemented in a separated internal class, for didactic purposes.";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("k", "10", "Number of candidate paths per demand, used as a source to compute the 1:1 pairs"));
		algorithmParameters.add(Triple.of("ea_numberIterations", "100", "Number of iterations (one per generation)"));
		algorithmParameters.add(Triple.of("ea_populationSize", "1000", "Number of elements in the population"));
		algorithmParameters.add(Triple.of("ea_offSpringSize", "500", "Number of childs in the offspring every generation"));
		algorithmParameters.add(Triple.of("ea_fractionParentsChosenRandomly", "0.5", "Fraction of the parents that are selected randomly for creating the offspring"));
		
		return algorithmParameters;
	}

	/**
	 * From a candidate path list, extract for each demand, the list of link-disjoint 
	 * path pairs. There pairs are candidates for becoming primary-backup path pairs for the demand.
	 */
	private Map<Long, List<Pair<Long, Long>>> computeCandidate11PairList(NetPlan netPlan, CandidatePathList cpl)
	{
		final Set<Long> demandIds = netPlan.getDemandIds();
		Map<Long, List<Pair<Long, Long>>> cpl11 = new LinkedHashMap<Long, List<Pair<Long, Long>>>();

		for (long demandId : demandIds)
		{
			cpl11.put(demandId, new ArrayList<Pair<Long, Long>>());
			Set<Long> pathIds = cpl.getDemandPaths(demandId);
			
			Iterator<Long> p1_it = pathIds.iterator();
			while(p1_it.hasNext())
			{
				final long p1 = p1_it.next();
				final List<Long> seqLinks_1 = cpl.getPathSequenceOfLinks(p1);
			
				Iterator<Long> p2_it = pathIds.iterator();
				while(p2_it.hasNext())
				{
					final long p2 = p2_it.next();
					if (p1 == p2) continue;
					
					final List<Long> seqLinks_2 = cpl.getPathSequenceOfLinks(p2);
					if (Collections.disjoint(seqLinks_1, seqLinks_2)) cpl11.get(demandId).add(Pair.of(p1, p2));
				}
			}

			if (cpl11.get(demandId).isEmpty()) throw new Net2PlanException("Demand " + demandId + " has no two link-disjoint paths");
		}

		return cpl11;
	}

	/**
	 * This class implements the core operations of the genetic algorithm to solve the network design problem. The problem consists of determining, for each
	 * network demand, one primary and backup path that are link disjoint. The backup path is assumed to carry no traffic, but when the primary fails.
	 * The optimization target is minimizing the average of two components. First, is the network congestion when no failure occurs in the network. The second is
	 * the worst network congestion when ONE link fails. With network congestion we mean the utilization in the bottleneck link.
	 * For each demand, we receive the set of admissible link-disjoint pairs to the problem. The problem solution is coded as an array with as many
	 * elements as demands. The d-th element, corresponds to the identifier of the 1:1 pair used as primary-backup for demand d.
	 *
	 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
	 * @version 1.0, May 2013
	 */
	public static class EvolutionaryAlgorithmCore
	{
		private final long[] linkIds, demandIds;
		private final Map<Long, Double> u_e;
		private final Map<Long, Double> h_d;
		private final Map<Long, List<Pair<Long, Long>>> potential11Pairs_d;
		private final CandidatePathList cpl;
		private final int D;
		private final int ea_numberIterations;
		private final int ea_populationSize;
		private final double ea_fractionChosenRandomly;
		private final int ea_offspringSize;
		private final Random rnd;
		private List<Map<Long, Integer>> population;
		private double[] costs;
		private double best_cost;
		private Map<Long, Integer> best_solution;

		/**
		 * Initialize the object with the problem data and the genetic algorithm specific parameters.
		 *
		 * @param u_e
		 * @param h_d
		 * @param potential11Pairs_d
		 * @param cpl
		 * @param ea_numberIterations
		 * @param ea_populationSize
		 * @param ea_fractionParentsChosenRandomly
		 * @param ea_offSpringSize
		 */
		public EvolutionaryAlgorithmCore(Map<Long, Double> u_e, Map<Long, Double> h_d, Map<Long, List<Pair<Long, Long>>> potential11Pairs_d, CandidatePathList cpl, int ea_numberIterations, int ea_populationSize, double ea_fractionParentsChosenRandomly, int ea_offSpringSize)
		{
			linkIds = LongUtils.toArray(u_e.keySet());
			demandIds = LongUtils.toArray(h_d.keySet());
			this.u_e = new LinkedHashMap(u_e);
			this.h_d = new LinkedHashMap(h_d);
			this.D = h_d.size();
			this.potential11Pairs_d = potential11Pairs_d;
			this.cpl = cpl;
			this.ea_numberIterations = ea_numberIterations;
			this.ea_populationSize = ea_populationSize;
			this.ea_fractionChosenRandomly = ea_fractionParentsChosenRandomly;
			this.ea_offspringSize = ea_offSpringSize;
			this.rnd = new Random();
		}

		/* Computes the cost of a given solution */
		private double computeCostSolution(Map<Long, Integer> solution_d)
		{
			/* Compute the congestion when there are no failures */
			Map<Long, Double> y_e = new LinkedHashMap<Long, Double>();
			for(long linkId : linkIds)
				y_e.put(linkId, 0.0);
				
			for (long demandId : demandIds)
			{
				final Pair<Long, Long> pathPair = potential11Pairs_d.get(demandId).get(solution_d.get(demandId));
				for (long linkId : cpl.getPathSequenceOfLinks(pathPair.getFirst()))
					y_e.put(linkId, y_e.get(linkId) + h_d.get(demandId));
			}

			double congestion_noFailure = 0;
			for(long linkId : linkIds)
			{
				double u_e_thisLink = u_e.get(linkId);
				double y_e_thisLink = y_e.get(linkId);
				double rho_thisLink = y_e_thisLink == 0 ? 0 : y_e_thisLink / u_e_thisLink;
				congestion_noFailure = Math.max(congestion_noFailure, rho_thisLink);
			}
			
			/* Compute the congestion for the single link failures */
			Map<Long, Double> congestion_s = new LinkedHashMap<Long, Double>(); /* congestion when fails link s */
			for (long linkId : linkIds)
			{
				congestion_s.put(linkId, 0.0);
				
				y_e.clear();
				for(long auxLinkId : linkIds)
					y_e.put(auxLinkId, 0.0);
				
				for (long demandId : demandIds)
				{
					final Pair<Long, Long> pathPair = potential11Pairs_d.get(demandId).get(solution_d.get(demandId));
					final long primaryPath = pathPair.getFirst();
					final long backupPath = pathPair.getSecond();
					final long usedPath = (cpl.linkBelongsToPath(linkId, primaryPath)) ? backupPath : primaryPath;
					for (long auxLinkId : cpl.getPathSequenceOfLinks(usedPath))
						y_e.put(auxLinkId, y_e.get(auxLinkId) + h_d.get(demandId));
				}

				for(long auxLinkId : linkIds)
				{
					double u_e_thisLink = u_e.get(auxLinkId);
					double y_e_thisLink = y_e.get(auxLinkId);
					double rho_thisLink = y_e_thisLink == 0 ? 0 : y_e_thisLink / u_e_thisLink;
					congestion_s.put(linkId, Math.max(congestion_s.get(linkId), rho_thisLink));
				}
			}
			
			double maxCongestion_s = 0;
			for(long linkId : linkIds)
				maxCongestion_s = Math.max(maxCongestion_s, congestion_s.get(linkId));

			return 0.5 * congestion_noFailure + 0.5 * maxCongestion_s;
		}

		/*
		 * Generates an initial population of solutions.
		 * Each solution is generated choosing randomly for each demand, one of the admissible 1:1 pairs
		 */
		private void generateInitialSolutions()
		{
			this.population = new ArrayList<Map<Long, Integer>>(ea_populationSize);
			this.costs = new double[ea_populationSize];

			for (int cont = 0; cont < ea_populationSize; cont++)
			{
				Map<Long, Integer> solution_d = new LinkedHashMap<Long, Integer>();
				for (long demandId : demandIds)
					solution_d.put(demandId, rnd.nextInt(this.potential11Pairs_d.get(demandId).size()));
				
				final double solutionCost = computeCostSolution(solution_d);
				population.add(solution_d);
				costs[cont] = solutionCost;
			}
		}

		/* Selects from the population ea_offspringSize*2 elements that will become parents, generating the offspring.
		 * A fraction of the parents (1-ea_fractionChosenRandomly) is chosen from the best in the population.
		 * A fraction ea_fractionChosenRandomly is chosen randomly from all the population, so that repetitions can occur */
		private LinkedList<Integer> operator_parentSelection()
		{
			final int numParentsToChoose = 2 * ea_offspringSize;
			final int numParentsChosenRandomly = (int) Math.floor(numParentsToChoose * ea_fractionChosenRandomly);
			final int numParentsChosenFromCost = numParentsToChoose - numParentsChosenRandomly;

			/* Initialize the list to be returned */
			LinkedList<Integer> chosenParents = new LinkedList<Integer>();

			/* Sorts the existing population: first the best cost */
			sortAscending(this.population, this.costs);

			/* Choose the best solutions as parents: as many as numParentsChosenFromCost */
			for (int cont = 0; cont < numParentsChosenFromCost; cont++)
			{
				chosenParents.add(cont);
			}

			/* The rest of the parents (numParentsChosenRandomly) are chosen randomly among the non-selected. Parents can be repeated */
			for (int cont = 0; cont < numParentsChosenRandomly; cont++)
			{
				chosenParents.add(rnd.nextInt(this.population.size()));
			}

			return chosenParents;
		}

		/* Receive a set of parents of size ea_offspringSize*2. From them, generates an offspring of size ea_offspringSize. Couples are generated randomly,
		 * each parent appears in only one couple. Given two parents, the descendant is created by, for each demand, picking from any parent (chosen randomly),
		 * the 1:1 pair to carry the traffic. */
		private List<Map<Long, Integer>> operator_crossover(LinkedList<Integer> parents)
		{
			/* The offspring to be returned */
			List<Map<Long, Integer>> offspring = new ArrayList<Map<Long, Integer>>(parents.size() / 2);

			/* Shuffle randomly the parent selection. Then, couples are formed randomly */
			Collections.shuffle(parents);

			/* Two consecutive parents are a couple */
			while (parents.size() >= 2)
			{
				final int firstParentId = parents.poll();
				final int secondParentId = parents.poll();

				final Map<Long, Integer> firstParent = population.get(firstParentId);
				final Map<Long, Integer> secondParent = population.get(secondParentId);

				/* The descendant chooses randomly for each demand, the paths from any of the two parents */
				Map<Long, Integer> descendant = new LinkedHashMap<Long, Integer>();
				for (long demandId : demandIds)
				{
					descendant.put(demandId, (rnd.nextBoolean()) ? firstParent.get(demandId) : secondParent.get(demandId));
				}

				offspring.add(descendant);
			}
			return offspring;
		}

		/*
		 * Receives a population (typically an offspring). For each element in
		 * the population, mutates the solution. A mutation consists of choosing
		 * a demand randomly, and changing randomly the 1:1 pair carrying the
		 * traffic for this demand (the same 1:1 pair as before could be chosen)
		 */
		private void operator_mutate(List<Map<Long, Integer>> offspring)
		{
			for(Map<Long, Integer> aux : offspring)
			{
				final long demandId = demandIds[rnd.nextInt(D)];
				final int number11PathPairs = this.potential11Pairs_d.get(demandId).size();
				aux.put(demandId, rnd.nextInt(number11PathPairs));
			}
		}

		/* Given the population from previous generation, and a new offspring, selects the elements that will pass to the next generation.
		 * These will be the best ea_populationSize among all the aggregated population */
		private void operator_select(List<Map<Long, Integer>> offspring)
		{
			/* Compute the costs of the offspring */
			double[] offspringCosts = new double[offspring.size()];
			for (int cont = 0; cont < offspring.size(); cont++)
			{
				offspringCosts[cont] = computeCostSolution(offspring.get(cont));
			}

			/* Append the offspring to the population */
			population.addAll(offspring);
			costs = DoubleUtils.concatenate(costs, offspringCosts);

			/* Reorder the new population */
			sortAscending(population, costs);

			/* Take the best ones, keeping the population size equal to ea_populationSize */
			while (population.size() > ea_populationSize)
			{
				population.remove(population.size() - 1);
			}

			costs = Arrays.copyOf(costs, ea_populationSize);
		}

		/**
		 * This method is called after creating the object, to launch the genetic algorithm. The method will first create an initial population. Then,
		 * will complete a number of iterations given by ea_numberIterations. In each iteration, a population is transformed using the evolutionary operators:
		 * parent selection, crossover, mutation, selection. The method keeps track of the best solution found, which is kept in the internal variables
		 * best_solution, best_cost. Then, they are publicly accessible when the method returns.
		 */
		public void evolve()
		{
			/* Generate the initial population */
			generateInitialSolutions();

			/* Update the best solution found so far */
			final int bestSolutionId = DoubleUtils.maxIndexes(this.costs, SearchType.FIRST)[0];
			best_cost = costs[bestSolutionId];
			best_solution = new LinkedHashMap<Long, Integer>(this.population.get(bestSolutionId));
			System.out.println("Initial population. Best solution cost: " + this.best_cost);

			/* Evolve: one iteration per generation */
			for (int it = 0; it < this.ea_numberIterations; it++)
			{
				LinkedList<Integer> parents = this.operator_parentSelection();
				List<Map<Long, Integer>> offspring = this.operator_crossover(parents);
				operator_mutate(offspring);
				operator_select(offspring);
				if (this.costs[0] < this.best_cost)
				{
					best_cost = this.costs[0];
					best_solution = new LinkedHashMap<Long, Integer>(this.population.get(0));
				}

				System.out.println("Iteration: " + it + ". Best solution cost: " + this.best_cost);
			}
		}

		/* Auxiliary function: Sorts a population in ascending order according to its cost (best cost first) */
		private void sortAscending(List<Map<Long, Integer>> population, double[] costs)
		{
			double[] copyCosts = DoubleUtils.copy(costs);
			List<Map<Long, Integer>> copyPopulation = new ArrayList<Map<Long, Integer>>(population);

			int[] orderedIndexes = DoubleUtils.sortIndexes(costs, OrderingType.ASCENDING);
			for (int newPosition = 0; newPosition < orderedIndexes.length; newPosition++)
			{
				final int oldIndex = orderedIndexes[newPosition];
				costs[newPosition] = copyCosts[oldIndex];
				population.set(newPosition, copyPopulation.get(oldIndex));
			}
		}
	}
}